# Expose a limited set of classes and functions so callers outside of
# the vcs package don't need to import deeper than `pip._internal.vcs`.
# (The test directory and imports protected by MYPY_CHECK_RUNNING may
# still need to import from a vcs sub-package.)
# Import all vcs modules to register each VCS in the VcsSupport object.
import stickybeak.vendored.pip._internal.vcs.bazaar
import stickybeak.vendored.pip._internal.vcs.git
import stickybeak.vendored.pip._internal.vcs.mercurial
import stickybeak.vendored.pip._internal.vcs.subversion  # noqa: F401
from stickybeak.vendored.pip._internal.vcs.versioncontrol import (  # noqa: F401
    RemoteNotFoundError,
    is_url,
    make_vcs_requirement_url,
    vcs,
)
